"""
Unit Tests for lmn transformations.
"""

import numpy as np
import unittest
import pyspedas
from pyspedas.tplot_tools import del_data, get_data, tnames, tplot, data_exists, tplot_restore
from pyspedas.projects.omni.omni_solarwind_load import omni_solarwind_load
from pyspedas.cotrans_tools.lmn_matrix_make import lmn_matrix_make
from pyspedas.projects.themis import fgm, state
from pyspedas.utilities.download import download


def _test_compare(tname1, tname2, center=False, approximate=False):
    """Compare two tplot variables."""
    try:
        data1 = get_data(tname1)
        t1 = data1.times
        y1 = data1.y
        data2 = get_data(tname2)
        t2 = data2.times
        y2 = data2.y

        if center:
            # Cut out the 10% of the edges of the data
            edge_cut = int(len(t1) * 0.1) - 1
            if edge_cut > 0 and 2 * edge_cut < len(t1):
                t1 = t1[edge_cut:-edge_cut]
                y1 = y1[edge_cut:-edge_cut]
                # find min and max of t1
                t1_min = np.min(t1)
                t1_max = np.max(t1)
                # for t2, find the indices of the values that are within t1_min and t1_max
                t2_indices = np.where((t2 >= t1_min) & (t2 <= t1_max))[0]
                t2 = t2[t2_indices]
                y2 = y2[t2_indices]

        do_match1 = np.allclose(t1, t2, rtol=1e-5, atol=1e-8)
        if approximate:
            do_match2 = np.allclose(y1, y2, rtol=1e-2, atol=1e-2)
        else:
            do_match2 = np.allclose(y1, y2, rtol=1e-5, atol=1e-8)
        do_match = do_match1 and do_match2
        if not do_match:
            print(f"Data for {tname1} and {tname2} do not match.")
            print(f"Data1: {data1.y}")
            print(f"Data2: {data2.y}")
    except Exception as e:
        print(f"Error comparing {tname1} and {tname2}: {e}")
        do_match = False
    return do_match


class TestLMNTransform(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Find the test data files to restore."""
        # Set to true when running locally
        localfile = False  
        # Set to your local directory, if running locally
        localdir = "C:/work/idl_working/"

        # The SPEDAS script that creates the file: general/tools/python_validate/lmn_python_validate.pro

        # Github files
        remote_server = "https://github.com/spedas/test_data/raw/refs/heads/main/"
        filename_solarwind = "solarwind_python_validate.tplot"
        filename_lmn = "lmn_python_validate.tplot"

        if localfile:
            cls.filename_sw = localdir + filename_solarwind
            cls.filename_lmn = localdir + filename_lmn
        else:
            remote_name = "cotrans_tools/" + filename_solarwind
            datafile = download(
                remote_file=remote_name,
                remote_path=remote_server,
                local_path="testdata",
                no_download=False,
            )
            if not datafile:
                # Skip tests
                raise unittest.SkipTest(
                    "Cannot download solarwind data validation file"
                )

            cls.filename_sw = datafile[0]

            remote_name = "cotrans_tools/" + filename_lmn
            datafile1 = download(
                remote_file=remote_name,
                remote_path=remote_server,
                local_path="testdata",
                no_download=False,
            )
            if not datafile1:
                # Skip tests
                raise unittest.SkipTest("Cannot download LMN data validation file")

            cls.filename_lmn = datafile1[0]

    # @unittest.skip("Skipping LMN tests")
    def test_lmn_1omni(self):
        """Test of OMNI variables."""
        # Load IDL variables from the test file
        del_data("*")
        tplot_restore(self.filename_sw)

        # Check if the IDL variables have been loaded
        tn = tnames()
        print(tn)
        idl_bz_def = "omni_BZ_idl_default"
        idl_p_def = "omni_P_idl_default"
        idl_bz_hro = "omni_BZ_idl_2020"
        idl_p_hro = "omni_P_idl_2020"
        idl_bz_hro2 = "omni_BZ_idl_2021"
        idl_p_hro2 = "omni_P_idl_2021"
        omni_names = [
            idl_bz_def,
            idl_p_def,
            idl_bz_hro,
            idl_p_hro,
            idl_bz_hro2,
            idl_p_hro2,
        ]
        for name in omni_names:
            self.assertTrue(data_exists(name), f"Data variable {name} does not exist.")

        # Load pyspedas data (default values) and compare
        # OMNI default values (constants)
        trange = ["2050-01-01", "2050-01-01 23:59:59"]
        omni_solarwind_load(trange=trange, level="hro", suffix="_py_default")
        t1 = _test_compare(idl_bz_def, "omni_solarwind_BZ_py_default")
        self.assertTrue(t1, f"Data for default BZ values do not match.")
        t2 = _test_compare(idl_p_def, "omni_solarwind_P_py_default")
        self.assertTrue(t2, f"Data for default P values do not match.")

        # OMNI HRO
        trange = ["2020-01-01", "2020-01-01 23:59:59"]
        omni_solarwind_load(trange=trange, level="hro", suffix="_py_2020")
        t3 = _test_compare(idl_bz_hro, "omni_solarwind_BZ_py_2020")
        self.assertTrue(t3, f"Data for 2020 HRO BZ values do not match.")
        t4 = _test_compare(idl_p_hro, "omni_solarwind_P_py_2020")
        self.assertTrue(t4, f"Data for 2020 HRO P values do not match.")

        # OMNI HRO2
        trange = ["2021-01-01", "2021-01-01 06:00:00"]
        omni_solarwind_load(trange=trange, level="hro2", suffix="_py_2021")
        t5 = _test_compare(idl_bz_hro2, "omni_solarwind_BZ_py_2021")
        self.assertTrue(t5, f"Data for 2021 HRO2 BZ values do not match.")
        t6 = _test_compare(idl_p_hro2, "omni_solarwind_P_py_2021")
        self.assertTrue(t6, f"Data for 2021 HRO2 P values do not match.")

    # @unittest.skip("Skipping LMN tests")
    def test_lmn_2lmn(self):
        """Test of LMN transform."""
        # Load LMN data
        del_data("*")
        tplot_restore(self.filename_lmn)

        # Check if the IDL LMN variables exist in tplot
        idl_pos = "tha_state_pos_gsm_idl"
        idl_b = "tha_fgl_gsm_idl"
        idl_hro = "tha_fgl_gsm_lmn_mat_hro_idl"
        idl_p = "OMNI_solarwind_P_idl"
        idl_bz = "OMNI_solarwind_BZ_idl"
        idl_names = [idl_pos, idl_b, idl_hro, idl_p, idl_bz]
        for name in idl_names:
            self.assertTrue(data_exists(name), f"Data variable does not exist: " + name)

        # Optionally plot the IDL data
        # tplot(idl_names)

        # Load pyspedas data
        trange = ["2022-01-01", "2022-01-01 06:00:00"]
        fgm(
            probe="a",
            level="l2",
            trange=trange,
            suffix="_py",
            time_clip=True,
            coord="gsm",
            get_support_data=True,
        )
        state(
            probe="a",
            trange=trange,
            suffix="_py",
            time_clip=True,
            get_support_data=True,
        )
        py_pos = "tha_pos_gsm_py"
        py_pos_i = py_pos + "_interpol"
        py_b = "tha_fgl_gsm_py"
        py_hro = "tha_fgl_gsm_lmn_mat_hro_py"
        py_p = "omni_solarwind_P"
        py_p_i = py_p + "_interpol"
        py_bz = "omni_solarwind_BZ"
        py_bz_i = py_bz + "_interpol"
        py_names = [py_pos, py_pos_i, py_b, py_hro, py_p, py_p_i, py_bz, py_bz_i]

        # Compute LMN
        lmn_matrix_make(py_pos, py_b, trange=trange, newname=py_hro)

        # Verify rotation matrix input and output coordinates
        in_coords, out_coords = pyspedas.rotmat_get_coords(py_hro)
        self.assertEqual(in_coords.upper(), "GSM")
        self.assertEqual(out_coords.upper(), "LMN")

        for name in py_names:
            self.assertTrue(data_exists(name), f"Data variable does not exist: " + name)
        # tplot(py_names)

        # Verify that position and B field as the same as in IDL
        t1 = _test_compare(idl_b, py_b, center=True)
        self.assertTrue(t1, f"B field data does not match.")
        t2 = _test_compare(idl_pos, py_pos, center=True)
        self.assertTrue(t2, f"Position data does not match.")

        # Verify that the interpolated data matches
        t3 = _test_compare(idl_p, py_p, center=True)
        self.assertTrue(t3, f"OMNI P data does not match.")
        t4 = _test_compare(idl_bz, py_bz, center=True)
        self.assertTrue(t4, f"OMNI Bz data does not match.")

        # Verify that the LMN matrices match
        t5 = _test_compare(idl_hro, py_hro, approximate=True)
        self.assertTrue(t5, f"LMN matrices do not match for HRO data.")
        
        print("test_lmn finished.")


if __name__ == "__main__":
    unittest.main()



import os

CONFIG = {'local_data_dir': 'cdaweb/',
          'cdas_endpoint': 'https://cdaweb.gsfc.nasa.gov/WS/cdasr/1/dataviews/sp_phys/',
          'remote_data_dir': 'https://cdaweb.gsfc.nasa.gov/sp_phys/data'}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'cdaweb/'])


import unittest

import pyspedas

import pyspedas
from pyspedas.projects.themis.analysis.scpot2dens import dens_pot
from pyspedas.projects.themis.analysis import scpot2dens
from pyspedas.tplot_tools import get_data
import numpy as np


class TestScpo2densValidation(unittest.TestCase):
    def test_dens_pot_values(self):
        """ Test supporting function dens_pot for initial values"""
        d1 = dens_pot(-1, 0)
        d2 = dens_pot(-1, 1)

        idl_d1 = 3048.9753
        idl_d2 = 34381.586

        self.assertAlmostEqual(d1, idl_d1, delta=0.1)
        self.assertAlmostEqual(d2, idl_d2, delta=0.1)

    def test_dens_pot_pder_values(self):
        """ Test supporting function dens_pot for derivative values"""
        _, d1 = dens_pot(-1, 0, True)
        _, d2 = dens_pot(-1, 1, True)

        idl_d1 = 6278.0029
        idl_d2 = 95702.727

        self.assertAlmostEqual(d1, idl_d1, delta=0.1)
        self.assertAlmostEqual(d2, idl_d2, delta=0.1)

    def test_dens_pot_list_values(self):
        """ Test supporting function dens_pot for tuple of values and derivative"""
        r1, d1 = dens_pot(np.array([0, -1]), 1, True)

        idl_r1 = np.array([3048.9753, 34381.586])
        idl_d1 = np.array([6278.0029, 95702.727])

        for p, idl in zip(r1, idl_r1):
            self.assertAlmostEqual(p, idl, delta=0.1)

        for p, idl in zip(d1, idl_d1):
            self.assertAlmostEqual(p, idl, delta=0.1)

    def test_scpot2dens_c(self):
        """
        Test satellite case of retrieving density with the change of the spacecraft potential offset.
        The offset if setup for probe c on (2007-7-20/17:24)
        """

        trange = ['2007-7-20/17:23:55', '2007-7-20/17:24:05']

        idl_Npot = np.array([19.2729, 53.6906, 53.0732])

        self.helper_scpot2dens('c', trange, idl_Npot)

    def test_scpot2dens_a(self):
        """
        Test satellite case of retrieving density using data from another spacecraft
        """

        trange = ['2012-01-01/18:10:05', '2012-01-01/18:10:10']

        idl_Npot = np.array([8.74609, 11.7515])

        self.helper_scpot2dens('a', trange, idl_Npot)

    def helper_scpot2dens(self, probe, trange, idl_Npot):
        """
        Helper function that evaluates scpot2dens for diffrent probes and tranges

        Parameters
        ----------
        probe: str
            THEMIS probe
        trange: list[str]
            trange string of the time tange
        idl_Npot: ndarray
            IDL values for the exact same call

        Returns
        -------
        None

        Notes
        -----
        The code that should be executed in IDL to obtain the  idl_Npot values is this:
        ```IDL
        trange = ['2012-01-01/18:10:05']
        probe = 'a'
        thm_load_esa,trange=trange, probe=probe, datat=' peer_avgtemp pe?r_density peer_sc_pot ', level=2
        ; Extract data:
        get_data,'th'+probe+'_peer_density',data=d
        dens_e= d.y
        dens_e_time= d.x
        ;
        get_data,'th'+probe+'_peir_density',data=d
        dens_i= d.y
        dens_i_time= d.x
        ;
        get_data,'th'+probe+'_peer_sc_pot',data=d
        sc_pot = d.y
        sc_pot_time = d.x
        ;
        get_data,'th'+probe+'_peer_avgtemp',data=d
        Te = d.y
        Te_time = d.x

        Npot = thm_scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, probe)
        print, Npot
        ```
        """

        pyspedas.projects.themis.esa(trange=trange, probe=probe,
                            varnames=[f'th{probe}_peer_density',
                                      f'th{probe}_peir_density',
                                      f'th{probe}_peer_sc_pot',
                                      f'th{probe}_peer_avgtemp'],
                            level='l2', time_clip=True)

        dens_e_time, dens_e = get_data(f'th{probe}_peer_density')
        dens_i_time, dens_i = get_data(f'th{probe}_peir_density')
        sc_pot_time, sc_pot = get_data(f'th{probe}_peer_sc_pot')
        Te_time, Te = get_data(f'th{probe}_peer_avgtemp')

        Npot = scpot2dens(sc_pot, sc_pot_time, Te, Te_time, dens_e, dens_e_time, dens_i, dens_i_time, probe=probe)

        for p, idl in zip(Npot, idl_Npot):
            self.assertAlmostEqual(p, idl, delta=0.1)

if __name__ == '__main__':
    unittest.main()



"""Test functions in the utilites folder."""
import unittest
import pyspedas
from copy import deepcopy

from pyspedas.tplot_tools import del_data, tplot, tplot_options, tplot_names
from pyspedas.tplot_tools import store_data, get_data
from pyspedas.tplot_tools import wildcard_expand, tname_byindex, tplot_wildcard_expand, tindex_byname

global_display=False
tp_data={'x':[1,2,3], 'y':[4,5,6]}

class UtilTestCases(unittest.TestCase):
    """Test functions related to tplot variable indexing and wildcards."""

    def test_wildcard_expand(self):
        # Test wildcard expansion against defined list
        master_list = ['Foo', 'foo', 'tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm']
        # Literal match
        result = wildcard_expand(master_list,"Foo")
        self.assertEqual(result,['Foo'])
        # Results should be ordered following patterns, not master list
        result = wildcard_expand(master_list,['foo', 'Foo'])
        self.assertEqual(result, ['foo', 'Foo'])
        # Wildcard expansion
        result = wildcard_expand(master_list,"th?_*_*")
        self.assertEqual(result,['tha_pos_gsm', 'thb_pos_gsm'])
        # Case sensitive (by default), no matches found
        result = wildcard_expand(master_list,"*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = wildcard_expand(master_list,"*SM", case_sensitive=False)
        self.assertEqual(result, ['tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm'])
        # Null pattern
        result = wildcard_expand(master_list, None)
        self.assertEqual(result, [])
        # Master list and pattern both null
        result = wildcard_expand(None,None)
        self.assertEqual(result, [])
        # Master list not a list (Formerly an error, now accepted)
        result = wildcard_expand('foo', 'foo')
        self.assertEqual(result, ['foo'])
        # Master list as space delimited string  (used to be an error, now accepted)
        master_list_str = 'Foo foo tha_pos_gsm thb_pos_gsm mms1_mec_r_sm'
        result = wildcard_expand(master_list_str, 'foo')
        self.assertEqual(result, ['foo'])
        # Master list and search list as space delimited strings
        result = wildcard_expand(master_list_str, 'foo bar')
        self.assertEqual(result, ['foo'])
        # Search list is a list containing some space delimited strings
        result = wildcard_expand(master_list_str, ['Foo','foo', 'thb_pos_gsm mms1_mec_r_sm'])
        self.assertEqual(result, ['Foo','foo', 'thb_pos_gsm', 'mms1_mec_r_sm'])
        # Master list is empty
        result = wildcard_expand([], 'foo')
        self.assertEqual(result, [])
        result = wildcard_expand([''], 'foo')
        self.assertEqual(result, [])
        # Patterns are empty
        result=wildcard_expand(master_list_str,[])
        self.assertEqual(result, [])
        result = wildcard_expand(master_list_str,'')
        self.assertEqual(result, [])
        # Do not split on spaces
        result=wildcard_expand(['foo bar','baz quux'], 'foo bar', split_whitespace=False)
        self.assertEqual(result, ['foo bar'])

    def test_wildcard_quiet(self):
        master = ['string1', 'string2', 'string3']
        patt1 = 'notfound'
        with self.assertLogs(level='WARNING') as log:
            # This should produce a warning
            result = wildcard_expand(master,patt1)
            self.assertTrue("No match found" in log.output[0])
            # With quiet==True, there should be no warning
            result = wildcard_expand(master,patt1,quiet=True)
            self.assertTrue(len(log.output) == 1)

    def test_tplot_wildcard_quiet(self):
        del_data('*')
        store_data('foo',deepcopy(tp_data))
        patt1 = 'notfound'
        with self.assertLogs(level='WARNING') as log:
            # This should produce a warning
            result = tplot_wildcard_expand(patt1)
            self.assertTrue("No match found" in log.output[0])
            # With quiet==True, there should be no warning
            result = tplot_wildcard_expand(patt1,quiet=True)
            self.assertTrue(len(log.output) == 1)
            del_data('*')
            # Should warn about empty list of tplot variable names
            result = tplot_wildcard_expand(patt1)
            self.assertTrue("is empty" in log.output[1])
            # len(log.output) is now 2
            # With quiet==True, there should be no warning
            result = tplot_wildcard_expand(patt1,quiet=True)
            self.assertTrue(len(log.output) == 2)



    def test_tplot_wildcard_expand(self):
        # Test wildcard expansion against current tplot variables
        del_data('*')
        master_list = ['Foo', 'foo', 'tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm']
        for var in master_list:
            store_data(var,deepcopy(tp_data))
        # Add a composite variable
        store_data('composite',data=['foo', 'tha_pos_gsm', 'thb_pos_gsm'])
        tplot_names()
        tn = tplot_names(quiet=True)
        print(tn)
        # Literal match
        result = tplot_wildcard_expand("Foo")
        self.assertEqual(result,['Foo'])
        # Results should be ordered following patterns, not master list
        result = tplot_wildcard_expand(['foo', 'Foo'])
        self.assertEqual(result, ['foo', 'Foo'])
        # Wildcard expansion
        result = tplot_wildcard_expand("th?_*_*")
        self.assertEqual(result,['tha_pos_gsm', 'thb_pos_gsm'])
        # Composite variable
        result = tplot_wildcard_expand('composite')
        self.assertEqual(result, ['composite'])
        # Case sensitive (by default), no matches found
        result = tplot_wildcard_expand("*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = tplot_wildcard_expand("*SM", case_sensitive=False)
        self.assertEqual(result, ['tha_pos_gsm', 'thb_pos_gsm', 'mms1_mec_r_sm'])
        # Mixed strings and integers with repeated values
        result = tplot_wildcard_expand(['Foo',1,2,2,'tha*'], case_sensitive=False)
        self.assertEqual(result, ['Foo','foo','tha_pos_gsm'])
        # Null pattern
        result = tplot_wildcard_expand(None)
        self.assertEqual(result, [])


    def test_tindex_byname(self):
        # Test lookups of tplot variable indices by name
        del_data('*')
        store_data('foo',data=deepcopy(tp_data))
        store_data('bar',data=deepcopy(tp_data))
        result=tindex_byname('foo')
        self.assertEqual(result,0)
        result=tindex_byname('bar')
        self.assertEqual(result,1)
        result=tindex_byname('doesnt_exist')
        self.assertEqual(result,None)

    def test_tname_byindex(self):
        # Test lookups of tplot variable names by indices
        del_data('*')
        # Out of bounds (no tplot variables yet)
        result=tname_byindex(0)
        self.assertEqual(result,None)
        store_data('foo',data=deepcopy(tp_data))
        store_data('bar',data=deepcopy(tp_data))
        result=tname_byindex(0)
        self.assertEqual(result,'foo')
        result=tname_byindex(1)
        self.assertEqual(result,'bar')
        # Index too large
        result=tname_byindex(2)
        self.assertEqual(result,None)
        # Negative index
        result=tname_byindex(-1)
        self.assertEqual(result,None)


    def test_plotting_wildcards_indices(self):
        # Test tplot using wildcards
        del_data('*')
        pyspedas.projects.themis.state(probe=['a', 'b', 'c', 'd', 'e'])
        # Wildcard pattern
        tplot_options('title',"Positions of THEMIS A B C D E")
        tplot('th?_pos', save_png='tplot_wildcard.png', display=global_display)
        # List of integer indices
        indices = [tindex_byname('tha_pos'), tindex_byname('thb_pos'), tindex_byname('thc_pos')]
        tplot_options('title',"Positions of THEMIS A B C")
        tplot(indices,save_png='tplot_byindex.png', display=global_display)
        # List including a bad name
        tplot_options('title',"Positions of THEMIS A B C D E")
        tplot(['th?_pos', 'doesnt_exist'], save_png='tplot_list_onebad.png', display=global_display)

    def test_tplot_names_embedded_spaces(self):
        # Tplot names can have embedded spaces, make sure they're treated correctly
        from pyspedas.tplot_tools import store_data, del_data, tnames, tplot_names, data_exists
        del_data('*')
        t=[1.0,2.0,3.0]
        d=[1,2,3]
        store_data('embedded spaces', data={'x':t, 'y':d})
        names=tnames()
        self.assertEqual(names, ['embedded spaces'])
        names=tplot_names()
        self.assertEqual(names, ['embedded spaces'])
        del_data('embedded spaces')
        self.assertFalse(data_exists('embedded spaces'))
        store_data('embedded spaces', data={'x':t, 'y':d})
        del_data('*')
        self.assertFalse(data_exists('embedded spaces'))
        store_data('embedded spaces', data={'x':t, 'y':d})
        # This should warn about 'd', 'e', and 'f' not found, but not warn about 'embedded' or 'spaces'
        vals = tplot_wildcard_expand(['embedded spaces', 'd e f'])
        self.assertEqual(vals,['embedded spaces'])


if __name__ == '__main__':
    unittest.main()



"""
Unit Tests for minvar function.
"""
import pyspedas

from pyspedas import fac_matrix_make
from pyspedas.tplot_tools import data_exists, tplot_names, tplot, get_data, tplot_copy, del_data, store_data
import numpy as np
import unittest
from pyspedas import tplot_restore
from numpy.testing import assert_allclose
from pyspedas import tvector_rotate, rotmat_get_coords
from pyspedas import download



class TestFac(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        #  Test tolerance
        cls.tol = 1e-10

        # Define a random data array
        rng = np.random.default_rng(seed=31415)
        cls.rdata = rng.random((10, 3))


        # Download tplot files
        # Comparison data was generated using both thm_fac_matrix_make and fac_matrix_make, using the script
        # general/tools/python_validate/fac_matrix_make_python_validate
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'cotrans_tools/fac_python_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path='testdata',
                            no_download=False)

        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        tplot_restore(filename)

        cls.fgm = get_data('thc_fgs_gse')
        tplot_copy('thc_fgs_gse', 'idl_thc_fgs_gse')
        cls.fgm_sm = get_data('thc_fgs_gse_sm601')
        tplot_copy('thc_fgs_gse_sm601', 'idl_thc_fgs_gse_sm601')
        cls.pos = get_data('thc_state_pos')
        tplot_copy('thc_state_pos', 'idl_thc_state_pos')

        cls.mat_xgse = get_data('thc_fgs_gse_sm601_fac_mat_xgse')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_xgse', 'idl_thc_fgs_gse_sm601_fac_mat_xgse')
        cls.mat_rgeo = get_data('thc_fgs_gse_sm601_fac_mat_rgeo')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_rgeo', 'idl_thc_fgs_gse_sm601_fac_mat_rgeo')
        cls.mat_phigeo = get_data('thc_fgs_gse_sm601_fac_mat_phigeo')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phigeo', 'idl_thc_fgs_gse_sm601_fac_mat_phigeo')
        cls.mat_mphigeo = get_data('thc_fgs_gse_sm601_fac_mat_mphigeo')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phism', 'idl_thc_fgs_gse_sm601_fac_mat_phism')
        cls.mat_phism = get_data('thc_fgs_gse_sm601_fac_mat_phism')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phism', 'idl_thc_fgs_gse_sm601_fac_mat_phism')
        cls.mat_mphism = get_data('thc_fgs_gse_sm601_fac_mat_mphism')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phism', 'idl_thc_fgs_gse_sm601_fac_mat_phism')
        cls.mat_ygsm = get_data('thc_fgs_gse_sm601_fac_mat_ygsm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_ygsm', 'idl_thc_fgs_gse_sm601_fac_mat_ygsm')

        cls.mat_xgse_thm = get_data('thc_fgs_gse_sm601_fac_mat_xgse_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_xgse_thm', 'idl_thc_fgs_gse_sm601_fac_mat_xgse_thm')
        cls.mat_rgeo_thm = get_data('thc_fgs_gse_sm601_fac_mat_rgeo_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_rgeo_thm', 'idl_thc_fgs_gse_sm601_fac_mat_rgeo_thm')
        cls.mat_phigeo_thm = get_data('thc_fgs_gse_sm601_fac_mat_phigeo_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phigeo_thm', 'idl_thc_fgs_gse_sm601_fac_mat_phigeo_thm')
        cls.mat_mphigeo_thm = get_data('thc_fgs_gse_sm601_fac_mat_mphigeo_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_mphigeo_thm', 'idl_thc_fgs_gse_sm601_fac_mat_mphigeo_thm')
        cls.mat_phism_thm = get_data('thc_fgs_gse_sm601_fac_mat_phism_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_phism_thm', 'idl_thc_fgs_gse_sm601_fac_mat_phism_thm')
        cls.mat_mphism_thm = get_data('thc_fgs_gse_sm601_fac_mat_mphism_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_mphism_thm', 'idl_thc_fgs_gse_sm601_fac_mat_mphism_thm')
        cls.mat_ygsm_thm = get_data('thc_fgs_gse_sm601_fac_mat_ygsm_thm')
        tplot_copy('thc_fgs_gse_sm601_fac_mat_ygsm_thm', 'idl_thc_fgs_gse_sm601_fac_mat_ygsm_thm')

        cls.vec_xgse = get_data('thc_fgs_fac_xgse')
        tplot_copy('thc_fgs_fac_xgse', 'idl_thc_fgs_fac_xgse')
        cls.vec_rgeo = get_data('thc_fgs_fac_rgeo')
        tplot_copy('thc_fgs_fac_rgeo', 'idl_thc_fgs_fac_rgeo')
        cls.vec_phigeo = get_data('thc_fgs_fac_phigeo')
        tplot_copy('thc_fgs_fac_phigeo', 'idl_thc_fgs_fac_phigeo')
        cls.vec_mphigeo = get_data('thc_fgs_fac_mphigeo')
        tplot_copy('thc_fgs_fac_mphigeo', 'idl_thc_fgs_fac_mphigeo')
        cls.vec_phism = get_data('thc_fgs_fac_phism')
        tplot_copy('thc_fgs_fac_phism', 'idl_thc_fgs_fac_phism')
        cls.vec_mphism = get_data('thc_fgs_fac_mphism')
        tplot_copy('thc_fgs_fac_mphism', 'idl_thc_fgs_fac_mphism')
        cls.vec_ygsm = get_data('thc_fgs_fac_ygsm')
        tplot_copy('thc_fgs_fac_ygsm', 'idl_thc_fgs_fac_ygsm')

        cls.vec_xgse_thm = get_data('thc_fgs_fac_xgse_thm')
        tplot_copy('thc_fgs_fac_xgse_thm', 'idl_thc_fgs_fac_xgse_thm')
        cls.vec_rgeo_thm = get_data('thc_fgs_fac_rgeo_thm')
        tplot_copy('thc_fgs_fac_rgeo_thm', 'idl_thc_fgs_fac_rgeo_thm')
        cls.vec_phigeo_thm = get_data('thc_fgs_fac_phigeo_thm')
        tplot_copy('thc_fgs_fac_phigeo_thm', 'idl_thc_fgs_fac_phigeo_thm')
        cls.vec_mphigeo_thm = get_data('thc_fgs_fac_mphigeo_thm')
        tplot_copy('thc_fgs_fac_mphigeo_thm', 'idl_thc_fgs_fac_mphigeo_thm')
        cls.vec_phism_thm = get_data('thc_fgs_fac_phism')
        tplot_copy('thc_fgs_fac_phism_thm', 'idl_thc_fgs_fac_phism_thm')
        cls.vec_mphism_thm = get_data('thc_fgs_fac_mphism')
        tplot_copy('thc_fgs_fac_mphism_thm', 'idl_thc_fgs_fac_mphism_thm')
        cls.vec_ygsm_thm = get_data('thc_fgs_fac_ygsm_thm')
        tplot_copy('thc_fgs_fac_ygsm_thm', 'idl_thc_fgs_fac_ygsm_thm')

        del_data('thc_*')



    def test_fac_xgse(self):
        """Test of other_dim = xgse"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='xgse',newname='mat_xgse')
        fac_in, fac_out = rotmat_get_coords('mat_xgse')
        self.assertEqual(fac_in, 'GSE')
        self.assertEqual(fac_out, 'FAC-XGSE')
        tvector_rotate('mat_xgse', 'idl_thc_fgs_gse', newname='vec_xgse')
        out_vec_coords = pyspedas.get_coords('vec_xgse')
        self.assertEqual(out_vec_coords,'FAC-XGSE')
        dat1 = get_data('mat_xgse')
        dat2 = get_data('vec_xgse')
        store_data('vecdiff',data={'x':dat2.times, 'y':self.vec_xgse.y - dat2.y})
        assert_allclose(dat1.y, self.mat_xgse.y, atol=1.0e-06)
        assert_allclose(dat2.y, self.vec_xgse.y, atol=1.0e-06)
        store_data('vecdiff',data={'x':dat2.times, 'y':self.vec_xgse.y - dat2.y})
        # Test rotmat_get_coords on variable that doesn't have the right metadata
        bogus_in, bogus_out = rotmat_get_coords('vecdiff')
        self.assertEqual(bogus_in, None)
        self.assertEqual(bogus_out, None)
        # Test tvector_rotate on a variable with no coords set
        tvector_rotate('mat_xgse', 'vecdiff', newname='bogus_vec_cotrans')
        #tplot('vec_xgse idl_thc_fgs_fac_xgse vecdiff')

    def test_fac_xgse_thm(self):
        """Test of other_dim = xgse"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='xgse',newname='mat_xgse')
        tvector_rotate('mat_xgse', 'idl_thc_fgs_gse', newname='vec_xgse')
        dat1 = get_data('mat_xgse')
        dat2 = get_data('vec_xgse')
        store_data('vecdiff',data={'x':dat2.times, 'y':self.vec_xgse_thm.y - dat2.y})
        assert_allclose(dat1.y, self.mat_xgse_thm.y, atol=1.0e-06)
        assert_allclose(dat2.y, self.vec_xgse_thm.y, atol=1.0e-06)
        store_data('vecdiff',data={'x':dat2.times, 'y':self.vec_xgse_thm.y - dat2.y})
        #tplot('vec_xgse idl_thc_fgs_fac_xgse vecdiff')

    def test_fac_rgeo(self):
        """Test of other_dim = xgse"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='rgeo',pos_var_name='idl_thc_state_pos', newname='mat_rgeo')
        tvector_rotate('mat_rgeo', 'idl_thc_fgs_gse', newname='vec_rgeo')
        dat1 = get_data('mat_rgeo')
        dat2 = get_data('vec_rgeo')
        assert_allclose(dat1.y, self.mat_rgeo.y, atol=.002)
        assert_allclose(dat2.y, self.vec_rgeo.y, atol=.002)

    def test_fac_rgeo_thm(self):
        """Test of other_dim = xgse"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='rgeo',pos_var_name='idl_thc_state_pos', newname='mat_rgeo')
        tvector_rotate('mat_rgeo', 'idl_thc_fgs_gse', newname='vec_rgeo')
        dat1 = get_data('mat_rgeo')
        dat2 = get_data('vec_rgeo')
        assert_allclose(dat1.y, self.mat_rgeo_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_rgeo_thm.y, atol=.002)

    def test_fac_phigeo(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='phigeo',pos_var_name='idl_thc_state_pos', newname='mat_phigeo')
        tvector_rotate('mat_phigeo', 'idl_thc_fgs_gse', newname='vec_phigeo')
        dat1 = get_data('mat_phigeo')
        dat2 = get_data('vec_phigeo')
        assert_allclose(dat1.y, self.mat_phigeo.y, atol=.002)
        assert_allclose(dat2.y, self.vec_phigeo.y, atol=.002)

    def test_fac_phigeo_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='phigeo',pos_var_name='idl_thc_state_pos', newname='mat_phigeo')
        tvector_rotate('mat_phigeo', 'idl_thc_fgs_gse', newname='vec_phigeo')
        dat1 = get_data('mat_phigeo')
        dat2 = get_data('vec_phigeo')
        assert_allclose(dat1.y, self.mat_phigeo_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_phigeo_thm.y, atol=.002)

    def test_fac_mphigeo(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='mphigeo',pos_var_name='idl_thc_state_pos', newname='mat_mphigeo')
        tvector_rotate('mat_mphigeo', 'idl_thc_fgs_gse', newname='vec_mphigeo')
        dat1 = get_data('mat_mphigeo')
        dat2 = get_data('vec_mphigeo')
        assert_allclose(dat1.y, self.mat_mphigeo.y, atol=.002)
        assert_allclose(dat2.y, self.vec_mphigeo.y, atol=.002)

    def test_fac_mphigeo_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='mphigeo',pos_var_name='idl_thc_state_pos', newname='mat_mphigeo')
        tvector_rotate('mat_mphigeo', 'idl_thc_fgs_gse', newname='vec_mphigeo')
        dat1 = get_data('mat_mphigeo')
        dat2 = get_data('vec_mphigeo')
        assert_allclose(dat1.y, self.mat_mphigeo_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_mphigeo_thm.y, atol=.002)

    def test_fac_phism(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='phism',pos_var_name='idl_thc_state_pos', newname='mat_phism')
        tvector_rotate('mat_phism', 'idl_thc_fgs_gse', newname='vec_phism')
        dat1 = get_data('mat_phism')
        dat2 = get_data('vec_phism')
        assert_allclose(dat1.y, self.mat_phism.y, atol=.002)
        assert_allclose(dat2.y, self.vec_phism.y, atol=.002)

    def test_fac_phism_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='phism',pos_var_name='idl_thc_state_pos', newname='mat_phism')
        tvector_rotate('mat_phism', 'idl_thc_fgs_gse', newname='vec_phism')
        dat1 = get_data('mat_phism')
        dat2 = get_data('vec_phism')
        assert_allclose(dat1.y, self.mat_phism_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_phism_thm.y, atol=.002)

    def test_fac_mphism(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='mphism',pos_var_name='idl_thc_state_pos', newname='mat_mphism')
        tvector_rotate('mat_mphism', 'idl_thc_fgs_gse', newname='vec_mphism')
        dat1 = get_data('mat_mphism')
        dat2 = get_data('vec_mphism')
        assert_allclose(dat1.y, self.mat_mphism.y, atol=.002)
        assert_allclose(dat2.y, self.vec_mphism.y, atol=.002)

    def test_fac_mphism_thm(self):
        """Test of other_dim = phigeo"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='mphism',pos_var_name='idl_thc_state_pos', newname='mat_mphism')
        tvector_rotate('mat_mphism', 'idl_thc_fgs_gse', newname='vec_mphism')
        dat1 = get_data('mat_mphism')
        dat2 = get_data('vec_mphism')
        assert_allclose(dat1.y, self.mat_mphism_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_mphism_thm.y, atol=.002)

    def test_fac_ygsm(self):
        """Test of other_dim = ygsm"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='ygsm',pos_var_name='idl_thc_state_pos', newname='mat_ygsm')
        tvector_rotate('mat_ygsm', 'idl_thc_fgs_gse', newname='vec_ygsm')
        dat1 = get_data('mat_ygsm')
        dat2 = get_data('vec_ygsm')
        assert_allclose(dat1.y, self.mat_ygsm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_ygsm.y, atol=.002)

    def test_fac_ygsm_thm(self):
        """Test of other_dim = ygsm"""

        fac_matrix_make('idl_thc_fgs_gse_sm601',other_dim='ygsm',pos_var_name='idl_thc_state_pos', newname='mat_ygsm')
        tvector_rotate('mat_ygsm', 'idl_thc_fgs_gse', newname='vec_ygsm')
        dat1 = get_data('mat_ygsm')
        dat2 = get_data('vec_ygsm')
        assert_allclose(dat1.y, self.mat_ygsm_thm.y, atol=.002)
        assert_allclose(dat2.y, self.vec_ygsm_thm.y, atol=.002)


if __name__ == '__main__':
    unittest.main()




import unittest

import numpy as np

from pyspedas.tplot_tools import data_exists
import pyspedas
from pyspedas.tplot_tools import time_double
from pyspedas.geopack import tt89
from pyspedas.geopack import tt96
from pyspedas.geopack import tt01
from pyspedas.geopack import tts04
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas.geopack.get_w_params import get_w
from pyspedas.geopack import trace_equator_89
from pyspedas.geopack import trace_iono_89
from pyspedas import tinterpol
from pyspedas import join_vec, store_data, get_data, tdeflag, del_data

trange = ['2015-10-16', '2015-10-17']

def gen_circle():
    import pyspedas
    # Generate a circle at 5 RE in the XZ plane
    angle = np.arange(0.0,361.0, 1.0)
    angle_rad = np.deg2rad(angle)
    y=np.zeros(len(angle_rad),np.float64)
    x=5.0*np.sin(angle_rad)*6371.2
    z=5.0*np.cos(angle_rad)*6371.2
    t=np.zeros(len(angle_rad))
    t[:] = pyspedas.time_double('2024-01-01/06:31:00') + np.arange(0.0,361.0,1.0)
    pos=np.zeros((len(angle_rad),3), np.float64)
    pos[:,0] = x
    pos[:,1] = y
    pos[:,2] = z
    pyspedas.store_data('circle_magpoles_5re',data={'x':t,'y':pos})
    pyspedas.set_coords('circle_magpoles_5re', 'GSM')

def get_params(model, g_variables=None):
    support_trange = [time_double(trange[0])-60*60*24, 
                      time_double(trange[1])+60*60*24]
    pyspedas.projects.kyoto.dst(trange=support_trange)
    pyspedas.projects.omni.data(trange=trange)
    join_vec(['BX_GSE', 'BY_GSM', 'BZ_GSM'])
    if model == 't01' and g_variables is None:
        g_variables = [6.0, 10.0]
    else:
        if g_variables is not None:
            if not isinstance(g_variables, str) and not isinstance(g_variables, np.ndarray):
                g_variables = None
    return get_tsy_params('kyoto_dst',
                    'BX_GSE-BY_GSM-BZ_GSM_joined',
                    'proton_density',
                    'flow_speed',
                    model,
                    pressure_tvar='Pressure',
                    g_variables=g_variables,
                    speed=True)


class LoadTestCases(unittest.TestCase):
    def test_igrf(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm', igrf_only=True)
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt89(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm')
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt96(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        params = get_params('t96')
        # This interpolation can result in NaNs in the position variable, so they need to be cleaned
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tdeflag('mms1_mec_r_gsm-itrp',newname='mms1_clean')
        tt96('mms1_clean', parmod=params)
        self.assertTrue(data_exists('mms1_clean_bt96'))

    def test_tt01(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        params = get_params('t01')
        # This can yield nans in the interpolated position variable for times outside the range of proton_density.
        # We don't want to pass NaNs to any of the geopack routines, so deflag
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tdeflag('mms1_mec_r_gsm-itrp',newname='mms1_clean')
        # No gvariables passed, defaults to two constants which will be replicated to the correct number of points
        tt01('mms1_clean', parmod=params)
        self.assertTrue(data_exists('mms1_clean_bt01'))
        del_data('mms1_clean_bt01')
        mec = get_data('mms1_clean')
        gvars = np.zeros((len(mec.times), 2))
        # Note that mec.times may now have a different number of points than proton_density!
        gvars[:, 0] = np.repeat(6.0, len(mec.times))
        gvars[:, 1] = np.repeat(10.0, len(mec.times))
        store_data('g_variables', data={'x': mec.times, 'y': gvars})
        params = get_params('t01', g_variables='g_variables')
        # Passing g_variables as tplot variable. It will get interpolated onto the correct number of points.
        tt01('mms1_clean', parmod=params)
        self.assertTrue(data_exists('mms1_clean_bt01'))
        del_data('mms1_clean_bt01')
        # Now we'll pass g_variables as an array. No interpolation will be done, so its size
        # needs to match proton_density, not mms1_clean!
        pd=get_data('proton_density')
        gvars = np.zeros((len(pd.times), 2))
        gvars[:, 0] = np.repeat(6.0, len(pd.times))
        gvars[:, 1] = np.repeat(10.0, len(pd.times))
        params = get_params('t01', g_variables=gvars)
        tt01('mms1_clean', parmod=params)
        self.assertTrue(data_exists('mms1_clean_bt01'))

    def test_tts04(self):
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        params = get_params('ts04')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tts04('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bts04'))

    def test_get_w(self):
        w_vals = get_w(trange=['2015-10-16', '2015-10-17'])

    def test_errors(self):
        # exercise some of the error code
        mec_vars = pyspedas.projects.mms.mec(trange=trange)
        params = get_params('ts04')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tts04('var_doesnt_exist')
        tts04('mms1_mec_r_gsm-itrp', parmod=None)
        tt01('var_doesnt_exist')
        tt01('mms1_mec_r_gsm-itrp', parmod=None)
        tt96('var_doesnt_exist')
        tt96('mms1_mec_r_gsm-itrp', parmod=None)
        tt89('var_doesnt_exist')
        invalidmodel = get_params('89')
        invalidg = get_params('t01', g_variables=1)
        invalidg = get_params('t01', g_variables='g_vars')
        notrange = get_w()  # no trange
        invalidtrange = get_w(trange=['2050-01-01', '2050-01-02'])

    def test_t89_equ_n(self):
        trace_equator_89(time_double('2007-03-23/00:00:00'),np.array([-2.0,0.0,1.0]),iopt=3)

    def test_t89_equ_s(self):
        trace_equator_89(time_double('2007-03-23/00:00:00'),np.array([-2.0,0.0,-1.0]),iopt=3)

    def test_t89_iono_n_n(self):
        trace_iono_89(time_double('2007-03-23/00:00:00'), np.array([-2.0,0.0,1.0]),iopt=3)

    def test_t89_iono_n_s(self):
        trace_iono_89(time_double('2007-03-23/00:00:00'), np.array([-2.0,0.0,1.0]),iopt=3,south=True)

    def test_t89_iono_s_n(self):
        trace_iono_89(time_double('2007-03-23/00:00:00'), np.array([-2.0,0.0,-1.0]),iopt=3)
    def test_t89_iono_s_s(self):
        trace_iono_89(time_double('2007-03-23/00:00:00'), np.array([-2.0,0.0,-1.0]),iopt=3, south=True)

    def test_t96_roi(self):
        import pyspedas
        gen_circle()
        params = get_params('t96')
        dat=get_data(params)
        circ_dat=get_data('circle_magpoles_5re')
        n=len(circ_dat.times)
        newdat=np.zeros((n,10), np.float64)
        newdat[:,:] = dat.y[0,]
        store_data(params,data={'x':circ_dat.times, 'y':newdat})
        tt96('circle_magpoles_5re', parmod=params)
        #pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bt96'])

    def test_t01_roi(self):
        import pyspedas
        gen_circle()
        params = get_params('t01')
        dat=get_data(params)
        circ_dat=get_data('circle_magpoles_5re')
        n=len(circ_dat.times)
        newdat=np.zeros((n,10), np.float64)
        newdat[:,:] = dat.y[0,]
        store_data(params,data={'x':circ_dat.times, 'y':newdat})
        tt01('circle_magpoles_5re', parmod=params)
        #pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bt01'])

    def test_t04_roi(self):
        import pyspedas
        gen_circle()
        params = get_params('ts04')
        dat=get_data(params)
        circ_dat=get_data('circle_magpoles_5re')
        n=len(circ_dat.times)
        newdat=np.zeros((n,10), np.float64)
        newdat[:,:] = dat.y[0,]
        store_data(params,data={'x':circ_dat.times, 'y':newdat})
        tts04('circle_magpoles_5re', parmod=params)
        #pyspedas.tplot(['circle_magpoles_5re','circle_magpoles_5re_bts04'])

if __name__ == '__main__':
    unittest.main()



"""Automated tests for tplot math and utility functions."""

import unittest
from pyspedas.tplot_tools import clean_spikes, smooth, tsmooth, tdotp, tcrossp, tnormalize, subtract_median, subtract_average, time_clip
from pyspedas.tplot_tools import get_data, store_data, time_double, time_string, time_datetime, del_data
from datetime import datetime,timezone
import numpy as np


class BaseTestCase(unittest.TestCase):
    """Data to be used in tests."""

    def setUp(self):
        """Create a tplot variable to be used in tests."""
        store_data('test', data={'x': [1., 2., 3., 4., 5., 6.],
                                 'y': [3., 5., 8., 15., 20., 1.]})


class AnalysisTestCases(BaseTestCase):
    """Test functions under tplot_math, and some other utilities."""

    def test_tdotp(self):
        store_data('var1', data={'x': [0], 'y': [[3, -3, 1]]})
        store_data('var2', data={'x': [0], 'y': [[4, 9, 2]]})
        dp = tdotp('var1', 'var2')
        dpdata = get_data('var1_dot_var2')
        self.assertTrue(dpdata.y == np.array([-13]))
        dp = tdotp('var1_doesnt_exist', 'var2')
        dp = tdotp('var1', 'var2_doesnt_exist')

    def test_tcrossp(self):
        """ cross product tests"""
        cp = tcrossp([3, -3, 1], [4, 9, 2], return_data=True)
        self.assertTrue(cp.tolist() == [-15, -2, 39])
        cp = tcrossp([3, -3, 1], [4, 9, 2])
        cp = get_data(cp)
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])
        store_data('var1', data={'x': [0], 'y': [[3, -3, 1]]})
        store_data('var2', data={'x': [0], 'y': [[4, 9, 2]]})
        cp = tcrossp('var1', 'var2', return_data=True)
        self.assertTrue(cp[0].tolist() == [-15, -2, 39])
        cp = tcrossp('var1', 'var2', newname='test_crossp')
        cp = get_data('test_crossp')
        self.assertTrue(cp.y[0, :].tolist() == [-15, -2, 39])

    def test_tnormalize(self):
        """ tests for normalizing tplot variables"""
        store_data('test_tnormalize', data={'x': [1, 2, 3, 4, 5], 'y': [[3, 2, 1], [1, 2, 3], [10, 5, 1], [8, 10, 14], [70, 20, 10]]})
        norm = tnormalize('test_tnormalize')
        normalized_data = get_data(norm)
        self.assertTrue(np.round(normalized_data.y[0, :], 4).tolist() == [0.8018, 0.5345, 0.2673])
        self.assertTrue(np.round(normalized_data.y[1, :], 4).tolist() == [0.2673, 0.5345, 0.8018])
        self.assertTrue(np.round(normalized_data.y[2, :], 4).tolist() == [0.8909, 0.4454, 0.0891])
        self.assertTrue(np.round(normalized_data.y[3, :], 4).tolist() == [0.4216, 0.527, 0.7379])
        self.assertTrue(np.round(normalized_data.y[4, :], 4).tolist() == [0.9526, 0.2722, 0.1361])

    def test_subtract_median(self):
        """Test subtract_median."""
        subtract_median('aaabbbccc')  # Test non-existent name
        subtract_median('test')
        d = get_data('test-m')
        self.assertTrue(d[1].tolist() == [-3.5, -1.5,  1.5,  8.5, 13.5, -5.5])
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., float('nan')]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        subtract_median('aaabbbcc')
        subtract_median('test1', newname='aabb')
        d = get_data('aabb')
        self.assertTrue(len(d[1]) == 6)
        subtract_median(['test', 'aabb'], newname='aaabbb')
        subtract_median('test1', overwrite=1)
        subtract_average('test', newname="testtest")
        subtract_average(['test-m', 'test'], newname="testtest2")

    def test_subtract_average(self):
        """Test subtract_average."""
        subtract_average('aaabbbccc')  # Test non-existent name
        subtract_average('test')
        d = get_data('test-d')
        self.assertTrue((np.round(d[1].tolist()) == [-6., -4., -1.,
                         6., 11., -8.]).all())
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., float('nan')]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        subtract_average('aaabbbcc')
        subtract_average('test1', newname='aabb')
        d = get_data('aabb')
        self.assertTrue(len(d[1]) == 6)
        subtract_average(['test', 'aabb'], newname='aaabbb')
        subtract_average('test1', overwrite=1)
        subtract_average('test1', newname="testtest")
        subtract_average(['test1', 'test'], newname="testtest2")


    def test_clean_spikes(self):
        """Test clean_spikes."""
        clean_spikes('aaabbbccc', nsmooth=3)  # Test non-existent name
        clean_spikes('test', nsmooth=3)
        d = get_data('test-despike')
        self.assertTrue(len(d[1]) == 6)
        # Now test 3 dim data.
        dn = [[3., 5., 8.], [15., 20., 1.], [3., 5., 8.], [15., 20., 1.],
              [23., 15., 28.], [15., 20., 1.]]
        store_data('test1', data={'x': [1., 2., 3., 4., 5., 6.], 'y': dn})
        clean_spikes('test1', nsmooth=3)
        d2 = get_data('test1-despike')
        clean_spikes('test', newname='test_desp', nsmooth=3, sub_avg=True)
        clean_spikes(['test', 'test1'], newname='test1-desp')
        clean_spikes('test1', overwrite=1)
        self.assertTrue(len(d2[1]) == 6)

    def test_tsmooth(self):
        """Test smooth."""
        tsmooth('aaabbbccc')  # Test non-existent name
        a = [1.0, 1.0, 2.0, 3.0, 4.0, 1.0, 4.0, 3.0, 2.0, 1.0, 1.0]
        x = smooth(a, 3)
        r = [1.0, 1.3333333333333333, 2.0, 3.0, 2.6666666666666665,
             3.0, 2.6666666666666665, 3.0, 2.0, 1.3333333333333333, 1.0]
        self.assertTrue(x == r)
        b = [1.0, 1.0, 2.0, 3.0, np.nan, np.nan, np.nan, np.nan, 2.0, 1.0, 1.0]
        y = smooth(b, width=3)
        ry = [1.0, 1.3333333333333333, 2.0, 1.6666666666666665, 1.0,
              np.nan, np.nan, 0.6666666666666666, 1.0, 1.3333333333333333, 1.0]
        self.assertTrue(y == ry)
        tsmooth('test')
        d = get_data('test-s')
        tsmooth('test', overwrite=1)
        tsmooth('test', newname="testtest")
        tsmooth(['test', 'test-s'], newname="testtest2")
        self.assertTrue(d[1].tolist() == [3.,  5.,  8., 15., 20.,  1.])

    def test_tplot_arithmetic(self):
        from pyspedas import add, subtract, multiply, divide, derive, tres
        from numpy.testing import assert_array_equal
        del_data('*')
        times = [0.0, 1.0, 2.0, 3.0, 4.0]
        dat1 = [1.0, 2.0, 3.0, 4.0, 5.0]
        dat2 = [10.0, 20.0, 30.0, 40.0, 50.0]
        store_data('v1',data={'x': times, 'y': dat1})
        store_data('v2',data={'x': times, 'y': dat2})
        add('v1', 'v2',newname='sum')
        subtract('v2','v1', newname='diff')
        multiply('v1', 'v2', newname='prod')
        divide('v2', 'v1', newname='quot')
        derive('v2',newname='ddx')
        sum_t, sum_d = get_data('sum')
        diff_t, diff_d = get_data('diff')
        prod_t, prod_d = get_data('prod')
        quot_t, quot_d = get_data('quot')
        ddx_t, ddx_d = get_data('ddx')
        assert_array_equal(diff_d, np.array([9.0, 18.0, 27.0, 36.0, 45.0]))
        assert_array_equal(sum_d,np.array([11.0, 22.0, 33.0, 44.0, 55.0]))
        assert_array_equal(quot_d, np.array([10.0, 10.0, 10.0, 10.0, 10.0]))
        assert_array_equal(prod_d, np.array([10.0, 40.0, 90.0, 160.0, 250.0]))
        assert_array_equal(ddx_d, np.array([10.0, 10.0, 10.0, 10.0, 10.0]))
        dt = tres('v1')
        self.assertEqual(dt, 1.0)

    def test_tplot_spectools(self):
        import pyspedas
        import numpy as np
        import math
        from numpy.testing import assert_array_equal
        del_data('*')
        # tpwrspc, pwrspc
        pyspedas.store_data('a', data={'x': range(100), 'y': np.random.random(100)})
        pyspedas.tpwrspc('a')
        spec = pyspedas.get_data('a_pwrspc')
        self.assertTrue(pyspedas.data_exists('a_pwrspc'))
        # pwr_spec
        time = [pyspedas.time_float("2020-01-01") + i for i in range(10000)]
        quantity = [math.sin(i) for i in range(10000)]
        pyspedas.store_data("dp", data={"x": time, "y": quantity})
        pyspedas.pwr_spec("dp", newname="dp_pwrspec")
        #pyspedas.tplot("dp_pwrspec")
        self.assertTrue(pyspedas.data_exists("dp_pwrspec"))
        # spec_mult
        pyspedas.store_data('h', data={'x':[0,4,8,12,16,19,21], 'y':[[8,1,1],[100,2,3],[4,2,47],[4,39,5],[5,5,99],[6,6,25],[7,-2,-5]],'v':[[1,1,50],[2,2,3],[100,4,47],[4,90,5],[5,5,99],[6,6,25],[7,7,-5]]})
        pyspedas.spec_mult('h')
        hdat = get_data('h_specmult')
        assert_array_equal(hdat.y, np.array([[8, 1, 50], [200, 4, 9], [400, 8, 2209], [16, 3510, 25], [25, 25, 9801], [36, 36, 625], [49, -14, 25]]))

    def test_add_across(self):
        import pyspedas
        from numpy.testing import assert_array_equal
        del_data('*')
        #Add across every column in the data
        pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1,50],[2,2,3],[100,4,47],[4,90,5],[5,5,99],[6,6,25],[7,7,-5]]})
        pyspedas.add_across('d',newname='d_aa')
        daa = get_data('d_aa')
        assert_array_equal(daa.y, np.array([[52], [7], [151], [99], [109], [37], [9]]))

        #Add across specific columns in the data
        pyspedas.store_data('b', data={'x':[2,5,8,11,14,17,20], 'y':[[1,1,1,1,1,1],[2,2,5,4,1,1],[100,100,3,50,1,1],[4,4,8,58,1,1],[5,5,9,21,1,1],[6,6,2,2,1,1],[7,7,1,6,1,1]]})
        pyspedas.add_across('b',column_range=[[1,2],[3,4]],newname='b_aap')
        dbaap = get_data('b_aap')
        assert_array_equal(dbaap.y,np.array([[2, 2], [7, 5], [103,  51], [12, 59], [14, 22], [8, 3], [8, 7]]))

    def test_crop(self):
        import pyspedas
        from numpy.testing import assert_array_equal
        del_data('*')
        pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        pyspedas.store_data('b', data={'x':[2,5,8,11,14,17,20], 'y':[[1,1,1,1,1,1],[2,2,5,4,1,1],[100,100,3,50,1,1],[4,4,8,58,1,1],[5,5,9,21,1,1],[6,6,2,2,1,1],[7,7,1,6,1,1]]})
        pyspedas.crop('a','b', replace=False)
        acrop_data = get_data('a_cropped')
        bcrop_data = get_data('b_cropped')
        assert_array_equal(acrop_data.times, np.array([ 4., 8., 12., 16.]))
        assert_array_equal(acrop_data.y, np.array([2, 3, 4, 5]))
        assert_array_equal(bcrop_data.times, np.array([2., 5., 8., 11., 14.]))
        assert_array_equal(bcrop_data.y, np.array([[1, 1, 1, 1, 1, 1], [2, 2, 5, 4, 1, 1], [100, 100, 3, 50, 1, 1], [4, 4, 8, 58, 1, 1], [5, 5, 9, 21, 1, 1]]
))
        # with replacement
        pyspedas.crop('a','b', replace=True)
        acrop_data = get_data('a')
        bcrop_data = get_data('b')
        assert_array_equal(acrop_data.times, np.array([ 4., 8., 12., 16.]))
        assert_array_equal(acrop_data.y, np.array([2, 3, 4, 5]))
        assert_array_equal(bcrop_data.times, np.array([2., 5., 8., 11., 14.]))
        assert_array_equal(bcrop_data.y, np.array([[1, 1, 1, 1, 1, 1], [2, 2, 5, 4, 1, 1], [100, 100, 3, 50, 1, 1], [4, 4, 8, 58, 1, 1], [5, 5, 9, 21, 1, 1]]))





if __name__ == '__main__':
    unittest.main()



"""
Tests for cotrans_tools functions.

These tests include the function in the following files:
    utilites/contrans.py
    utilities/cotrans_lib.py
    utilites/igrf.py
    utilites/j2000.py
    themis/cotrans_tools/dsl2gse.py

"""
import numpy as np
from numpy.testing import assert_allclose
import unittest
import pyspedas
import logging
from pyspedas.projects.themis.cotrans.dsl2gse import dsl2gse
from pyspedas.cotrans_tools.cotrans import cotrans
from pyspedas.cotrans_tools.fac_matrix_make import fac_matrix_make
from pyspedas.tplot_tools import get_data, store_data, del_data, get_coords, set_coords
from pyspedas import sm2mlt, tplot_copy, set_units, get_units

global_display=False

class CotransTestCases(unittest.TestCase):
    """Tests for cotrans."""

    def test_fac_matrix_make(self):
        with self.assertLogs(level='ERROR') as log:
            doesntexist = fac_matrix_make('doesnt_exist')
            self.assertTrue(doesntexist is None)
            self.assertIn("Error reading tplot variable: doesnt_exist", log.output[0])

    def test_get_set_coords(self):
        """ Test for pytplot.set_coords/get_coords """
        from pyspedas.tplot_tools import set_coords, get_coords

        del_data()
        doesntexist = get_coords('test_coord')
        self.assertTrue(doesntexist is None)   
        store_data('test_coord', data={'x': [1, 2, 3, 4, 5], 'y': [1, 1, 1, 1, 1]})
        with self.assertLogs(level='ERROR') as log:
            cotrans(name_in='test_coord', coord_out="geo")
            self.assertIn("cotrans error: No input coordinates were provided.", log.output[0])
        before = get_coords('test_coord')
        self.assertTrue(before is None)
        setcoord = set_coords('test_coord', 'GSE')
        self.assertTrue(setcoord)
        after = get_coords('test_coord')
        self.assertTrue(after == 'GSE')
        md = get_data('test_coord', metadata=True)
        md['data_att']['units'] = 'km'
        setcoord = set_coords('test_coord', 'GSM')
        self.assertTrue(setcoord)
        md_after = get_data('test_coord', metadata=True)
        after = get_coords('test_coord')
        self.assertTrue(after == 'GSM')
        self.assertTrue(md_after['data_att']['units'] == 'km')
        setcoord = set_coords('doesnt_exist', 'GSM')

    def test_get_set_units(self):
        """ Test for pytplot.set_coords/get_coords """
        from pyspedas.tplot_tools import set_units, get_units, set_coords, get_coords

        del_data()
        doesntexist = get_units('test_units')
        self.assertTrue(doesntexist is None)
        store_data('test_units', data={'x': [1, 2, 3, 4, 5], 'y': [1, 1, 1, 1, 1]})
        before = get_units('test_units')
        self.assertTrue(before is None)
        setunits = set_units('test_units', 'Km')
        self.assertTrue(setunits)
        after = get_units('test_units')
        self.assertTrue(after == 'Km')
        set_coords('test_units', 'GEO')
        setunits = set_units('test_units', 'mm')
        self.assertTrue(setunits)
        coords_after = get_coords('test_units')
        units_after = get_units('test_units')
        self.assertTrue(coords_after == 'GEO')
        self.assertTrue(units_after == 'mm')

    def test_dsl2gse(self):
        """Test themis.cotrans_tools.dsl2gse."""
        del_data()
        # Try with missing variables. It should exit without problems.
        with self.assertLogs(level='ERROR') as log:
            dsl2gse('tha_fgl_dsl', 'tha_fgl_gse')
            self.assertIn("Variables needed: ['tha_fgl_dsl']", log.output[0])
            self.assertIn("Variables missing: ['tha_fgl_dsl']", log.output[1])
            self.assertIn("Please load missing variables.", log.output[2])
        # Now load the needed variables.
        time_range = ['2017-03-23 00:00:00', '2017-03-23 23:59:59']
        pyspedas.projects.themis.state(probe='a', trange=time_range,
                              get_support_data=True,
                              varnames=['tha_spinras', 'tha_spindec'])
        pyspedas.projects.themis.fgm(probe='a', trange=time_range,
                            varnames=['tha_fgl_dsl'])

        fac_matrix_make('tha_fgl_dsl')

        dsl2gse('tha_fgl_dsl', 'tha_fgl_gse')

        t, d = get_data('tha_fgl_gse')
        # Now test the inverse.
        dsl2gse('tha_fgl_dsl', 'tha_fgl_gse',
                isgsetodsl=True)

        self.assertTrue(abs(d[0].tolist()[0]-15.905078404701147) <= 1e-6)
        self.assertTrue(abs(d[0].tolist()[1]--13.962618931740064) <= 1e-6)
        self.assertTrue(abs(d[0].tolist()[2]-16.392516225582813) <= 1e-6)

        self.assertTrue(abs(d[50000].tolist()[0]-16.079111468932435) <= 1e-6)
        self.assertTrue(abs(d[50000].tolist()[1]--18.858874541698583) <= 1e-6)
        self.assertTrue(abs(d[50000].tolist()[2]-14.75796300561617) <= 1e-6)

    def test_cotrans(self):
        """Test cotrans.py GEI->GEO with Themis data."""
        del_data()
        trange = ['2010-02-25/00:00:00', '2010-02-25/23:59:59']
        probe = 'a'
        name_in = "tha_pos"
        name_out = "tha_pos_new_geo"
        pyspedas.projects.themis.state(probe=probe, trange=trange,
                              time_clip=True, varnames=[name_in])
        cotrans(name_in=name_in, name_out=name_out,
                coord_in="gei", coord_out="geo")
        din = get_data(name_in)
        in_len = len(din[0])
        dout = get_data(name_out)
        out_len = len(dout[0])
        cotrans(name_in=name_in, coord_in="gei", coord_out="geo")
        self.assertTrue(out_len == in_len)

    def test_cotrans_coord_mismatch(self):
        """Test that cotrans rejects a request where in_coord does not match the system from the variable metadata."""
        del_data()
        trange = ['2010-02-25/00:00:00', '2010-02-25/23:59:59']
        probe = 'a'
        name_in = "tha_pos"
        name_out = "tha_pos_new_geo"
        pyspedas.projects.themis.state(probe=probe, trange=trange,
                              time_clip=True, varnames=[name_in])
        # Metadata coordinate system is GEI, but requesting GSM->GEO transform.  This should generate an error message
        # and return failure.
        with self.assertLogs(level='ERROR') as log:
            result = cotrans(name_in=name_in, name_out=name_out,
                            coord_in="gsm", coord_out="geo")
            self.assertTrue(result == 0)

    def test_cotrans_igrf(self):
        """Test GSE->GSM and IGRF."""
        del_data()
        d = [[245.0, -102.0, 251.0], [775.0, 10.0, -10],
             [121.0, 545.0, -1.0], [304.65, -205.3, 856.1],
             [464.34, -561.55, -356.22]]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(name_out=name_out, time_in=t, data_in=d,
                coord_in="gse", coord_out="gsm")
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        gsm = dout[1][1]
        res = [775.0, 11.70357713,-7.93890939]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(gsm[0]-res[0]) <= 1e-6)
        self.assertTrue(abs(gsm[1]-res[1]) <= 1e-6)
        self.assertTrue(abs(gsm[2]-res[2]) <= 1e-6)

    def test_cotrans_j2000(self):
        """Test GEI->J2000 and J2000 params."""
        del_data()
        d = [[245.0, -102.0, 251.0], [775.0, 10.0, -10],
             [121.0, 545.0, -1.0], [304.65, -205.3, 856.1],
             [464.34, -561.55, -356.22]]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(name_out=name_out, time_in=t, data_in=d,
                coord_in="gei", coord_out="j2000")
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        j2000 = dout[1][1]
        res = [775.01595209, 6.59521058, -11.47942543]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(j2000[0]-res[0]) <= 1e-6)
        self.assertTrue(abs(j2000[1]-res[1]) <= 1e-6)
        self.assertTrue(abs(j2000[2]-res[2]) <= 1e-6)

    def test_cotrans_geigsm(self):
        """Test daisy chain of multiple transofrmations, GEI->GSE->GSM."""
        del_data()
        d = [[245.0, -102.0, 251.0], [775.0, 10.0, -10],
             [121.0, 545.0, -1.0], [304.65, -205.3, 856.1],
             [464.34, -561.55, -356.22]]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        name_out = "tname_out"
        cotrans(name_out=name_out, time_in=t, data_in=d,
                coord_in="gei", coord_out="gsm")
        in_len = len(t)
        dout = get_data(name_out)
        out_len = len(dout[0])
        gsm = dout[1][1]
        res = [ 45.98475132, 762.29816225, 132.7098884 ]
        self.assertTrue(out_len == in_len)
        self.assertTrue(abs(gsm[0]-res[0]) <= 1e-6)
        self.assertTrue(abs(gsm[1]-res[1]) <= 1e-6)
        self.assertTrue(abs(gsm[2]-res[2]) <= 1e-6)

    def test_all_cotrans(self):
        """Test all cotrans pairs.

        Apply transformation, then inverse transformation and compare.
        """
        with self.assertLogs(level='ERROR') as log:
            cotrans()
            self.assertIn("cotrans error: No output coordinates were provided.", log.output[0])
        all_cotrans = ['gei', 'geo', 'j2000', 'gsm', 'mag', 'gse', 'sm']
        d = [[245.0, -102.0, 251.0], [775.0, 10.0, -10],
             [121.0, 545.0, -1.0], [304.65, -205.3, 856.1],
             [464.34, -561.55, -356.22]]
        dd1 = d[1]
        t = [1577112800, 1577308800, 1577598800, 1577608800, 1577998800]
        in_len = len(t)
        name1 = "name1"
        name2 = "name2"
        count = 0
        # Test non-existent systems.
        with self.assertLogs(level='ERROR') as log:
            result = cotrans(name_out=name1, time_in=t, data_in=d,
                            coord_in="badcoord", coord_out="gei")
            self.assertTrue(result == 0)
            self.assertIn("cotrans error: Requested input coordinate system", log.output[0])
            result = cotrans(name_out=name1, time_in=t, data_in=d,
                            coord_in="gei", coord_out="badcoord")
            self.assertTrue(result == 0)
            self.assertIn("cotrans error: Requested output coordinate system", log.output[1])

        # Test empty data.
        with self.assertLogs(level='ERROR') as log:
            cotrans(name_out=name1, time_in=t, data_in=[],
                    coord_in="gei", coord_out="geo")
            cotrans(time_in=t, data_in=d, coord_in="gse", coord_out="gsm")
            self.assertIn("cotrans error: Data is empty.", log.output[0])
            self.assertEqual(len(log.output), 1)
        # Test all combinations.
        with self.assertLogs(level='WARNING') as log:
            for i, coord_in in enumerate(all_cotrans):
                for j, coord_out in enumerate(all_cotrans):
                    count += 1
                    del_data()
                    cotrans(name_out=name1, time_in=t, data_in=d,
                            coord_in=coord_in, coord_out=coord_out)
                    dout = get_data(name1)
                    out_len1 = len(dout[0])
                    self.assertTrue(out_len1 == in_len)
                    # Now perform inverse transformation.
                    cotrans(name_in=name1, name_out=name2,
                            coord_in=coord_out, coord_out=coord_in)
                    dout2 = get_data(name2)
                    out_len2 = len(dout2[0])
                    dd2 = dout2[1][1]
                    logging.info("%d --- in: %s out: %s", count, coord_in, coord_out)
                    # print(dout[1][1])
                    # print(dd2)
                    self.assertTrue(out_len2 == in_len)
                    self.assertTrue(abs(dd1[0]-dd2[0]) <= 1e-6)
                    self.assertTrue(abs(dd1[1]-dd2[1]) <= 1e-6)
                    self.assertTrue(abs(dd1[2]-dd2[2]) <= 1e-6)
                    if coord_out == coord_in:
                        self.assertIn("Warning: coord_in equal to coord_out.", log.output[i*2])
                        self.assertIn("Warning: coord_in equal to coord_out.", log.output[j*2+1])
                

    def test_mlt(self):
        '''Test sm2mlt.

        Data is from goes18, 2023-01-01. [0, 100, 600, 1000]
        '''

        # SM coordinates and MLT values from IDL
        sm_idl = [[33199.660, 25815.376, 2995.6487], [18814.747, 37614.243, 2990.7136],
                  [-40428.563, -11607.452, 2995.3227], [15798.372, -38974.509, 2995.0028]]
        mlt_idl = [14.524527, 16.228377, 1.0679544, 7.4710163]

        x = [item[0] for item in sm_idl]
        y = [item[1] for item in sm_idl]
        z = [item[2] for item in sm_idl]

        mlt_python = sm2mlt(x, y, z)

        for i in range(4):
            self.assertTrue(abs(mlt_idl[i]-mlt_python[i]) <= 1e-6)

    def test_units_coords(self):
        '''Test setting units and coordinate systems (including wildcards)
        '''

        test_dat = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        test_times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])

        store_data('v1', data={'x':test_times, 'y':test_dat})
        tplot_copy('v1', 'v2')
        tplot_copy('v1', 'v3')
        tplot_copy('v1','v4')

        set_coords('v?', 'GSE')
        c1 = get_coords('v1')
        c2 = get_coords('v2')
        c3 = get_coords('v3')
        c4 = get_coords('v4')
        self.assertTrue(c1 == 'GSE')
        self.assertTrue(c2 == 'GSE')
        self.assertTrue(c3 == 'GSE')
        self.assertTrue(c4 == 'GSE')

        set_units('v?', 'nT')
        u1 = get_units('v1')
        u2 = get_units('v2')
        u3 = get_units('v3')
        u4 = get_units('v4')
        self.assertTrue(u1 == 'nT')
        self.assertTrue(u2 == 'nT')
        self.assertTrue(u3 == 'nT')
        self.assertTrue(u4 == 'nT')

        res = set_coords('nonexistent', 'GSE')
        self.assertFalse(res)

        res = set_units('xxx yyy zzz', 'nT')
        self.assertFalse(res)

    def test_geogsm_geopack(self):
        # Originally reported by Bob Weigel -- PySPEDAS GEO->GSM cotrans had weird artifacts at year boundaries
        # compared to Geopack GEO->GSM transform and other Python packages
        # The SPEDAS script that creates the file: general/tools/python_validate/cotrans_gsm_test.pro
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'cotrans_tools/geogsm_gp_test.tplot'
        datafile = pyspedas.download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path='testdata',
                            no_download=False)

        pyspedas.tplot_restore(datafile[0])
        pyspedas.tplot_names()
        pyspedas.cotrans('zgeo','zgeo_gsm_py',coord_in='geo',coord_out='gsm')
        pyspedas.options('zgeo_gsm_py','ytitle','zgeo_gsm_py')
        pyspedas.subtract('zgeo_gsm_py','zgeo_gsm',newname='pydiff')
        pyspedas.options('pydiff','ytitle','idl_python_diff')
        pyspedas.subtract('zgeo_gsm_gp','zgeo_gsm',newname='gpdiff')
        pyspedas.options('gpdiff','ytitle','idl_geopack_diff')
        pyspedas.tplot(['zgeo', 'zgeo_gsm', 'zgeo_gsm_py', 'pydiff', 'gpdiff'],display=global_display,save_png='geo_gsm_diffs.png')
        didl=get_data('zgeo_gsm')
        dpy = get_data('zgeo_gsm_py')
        # Max relative difference around 2.7e-07
        assert_allclose(didl.y, dpy.y, atol=3.0e-07)

if __name__ == '__main__':
    unittest.main()



"""
Unit Tests for minvar function.
"""
import pyspedas
from pyspedas import minvar
from pyspedas import minvar_matrix_make
from pyspedas import rotmat_get_coords, get_coords
from pyspedas.tplot_tools import data_exists, tplot_names, tplot, get_data, tplot_copy, del_data, store_data
import numpy as np
import unittest
from pyspedas import tplot_restore
from numpy.testing import assert_allclose
from pyspedas import tvector_rotate
from pyspedas import download



class TestMinvar(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        #  Test tolerance
        cls.tol = 1e-10

        # Define a random data array
        rng = np.random.default_rng(seed=31415)
        cls.rdata = rng.random((10, 3))

        # The SPEDAS script that creates the file: general/tools/python_validate/minvar_python_validate.pro

        # Download tplot files
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'cotrans_tools/mva_python_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path='testdata',
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        tplot_restore(filename)

        cls.fgm = get_data('thb_fgs_gsm_mvaclipped1')
        cls.mat = get_data('thb_fgs_gsm_mvaclipped1_mva_mat')
        cls.vals = get_data('mva_vals')
        cls.min = get_data('mva_min')
        cls.int = get_data('mva_int')
        cls.max = get_data('mva_max')
        cls.rot = get_data('thb_fgs_gsm_mvaclipped1_rot')
        tplot_copy('thb_fgs_gsm_mvaclipped1','idl_thb_fgs_gsm_mvaclipped1')
        tplot_copy('thb_fgs_gsm_mvaclipped1_mva_mat','idl_thb_fgs_gsm_mvaclipped1_mva_mat')
        tplot_copy('mva_min','idl_mva_min')
        tplot_copy('mva_int','idl_mva_int')
        tplot_copy('mva_max','idl_mva_max')
        tplot_copy('mva_vals','idl_mva_vals')
        tplot_copy('thb_fgs_gsm_mvaclipped1_rot','idl_thb_fgs_gsm_mvaclipped1_rot')

        cls.fgm2 = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar')
        cls.mat2 = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat')
        cls.inp2 = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_bp')
        cls.vals2 = get_data('mva_vals2')
        cls.min2 = get_data('mva_min2')
        cls.int2 = get_data('mva_int2')
        cls.max2 = get_data('mva_max2')
        cls.rot2 = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot')
        tplot_copy('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar','idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar')
        tplot_copy('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat','idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat')
        tplot_copy('mva_min2','idl_mva_min2')
        tplot_copy('mva_int2','idl_mva_int2')
        tplot_copy('mva_max2','idl_mva_max2')
        tplot_copy('mva_vals2','idl_mva_vals2')
        tplot_copy('mms1_fgm_b_gsm_srvy_l2_bvec_bp','idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp')
        tplot_copy('mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot','idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot')

        del_data('thb_*')
        del_data('mms1_*')
        del_data('mva*')



    def test_minvar_basic(self):
        """Test of basic input and output"""

        # Synthetic data of zeros
        data = np.zeros([2, 3])
        vrot, v, w = minvar(data)
        self.assertTrue(np.sum(vrot - data) < self.tol)
        self.assertTrue(np.sum(v - np.diag(np.ones(3))) < self.tol)
        self.assertTrue(np.sum(w - np.zeros(3)) < self.tol)

    def test_minvar_rotation(self):
        """Test of the rotation matrix"""
        vrot, v, w = minvar(self.rdata)
        # Determinant of rotation matrix should be = 1
        self.assertTrue((np.linalg.det(v) - 1) < self.tol)

    def test_minvar_total(self):
        """Test of same square root of total of squares """
        vrot, v, w = minvar(self.rdata)
        total1 = (self.rdata**2).sum(axis=1)
        total2 = (vrot ** 2).sum(axis=1)
        # Compare totals
        self.assertTrue(np.sum(total1 - total2) < self.tol)

    def test_minvar_code_coverage(self):
        """Test to cover the code from IDL"""
        data = np.array([[0, 0, 1], [0, 0, 1]])
        vrot, v, w = minvar(data)
        # Case of np.sum(w) == 0.0
        self.assertTrue(w.sum() < self.tol)

        # This should be not Right Handed (?...)
        data = np.array([[0, -1, 1], [-1, -1, 1]])
        # case if YcrossZdotX < 0
        vrot, v, w = minvar(data)
        YcrossZdotX = v[0, 0] * (v[1, 1] * v[2, 2] - v[2, 1] * v[1, 2])
        # YcrossZdotX Should be positive after that
        self.assertTrue(YcrossZdotX > 0)

        # should tigger case if v[2, 2] < 0: (?...)
        data = np.array([[-0.1, -0.9, 0.5], [-1, 1, -0.9]])
        vrot, v, w = minvar(data)
        # v[2,2] Should be positive after that
        self.assertTrue(v[2, 2] > 0)

    def test_minvar_matrix_make_day(self):
        trange=['2007-07-10', '2007-07-11']
        pyspedas.projects.themis.fgm(probe='c', trange=trange, level='l2', coord='gse')
        minvar_matrix_make('thc_fgs_gse',tstart='2007-07-10/07:54:00',tstop='2007-07-10/07:56:30')
        self.assertTrue(data_exists('thc_fgs_gse_mva_mat'))
        mat_coord=rotmat_get_coords('thc_fgs_gse_mva_mat')
        pyspedas.tvector_rotate('thc_fgs_gse_mva_mat','thc_fgs_gse',newname='mva_data_day')
        self.assertTrue((data_exists('mva_data_day')))
        out_coord=get_coords('mva_data_day')
        self.assertTrue(len(mat_coord) == 2)
        self.assertEqual(mat_coord[0].lower(),"gse")
        self.assertEqual(mat_coord[1].lower(),"mva")
        self.assertEqual(out_coord.lower(),'mva')

    def test_minvar_matrix_extra_outputs(self):
        trange=['2007-07-10', '2007-07-11']
        pyspedas.projects.themis.fgm(probe='c', trange=trange, level='l2', coord='gse')
        minvar_matrix_make('thc_fgs_gse',tstart='2007-07-10/07:54:00',tstop='2007-07-10/07:56:30',
                           evname='mva_lambdas',tminname='mva_min',tmidname='mva_mid',tmaxname='mva_max')
        self.assertTrue(data_exists('thc_fgs_gse_mva_mat'))
        pyspedas.tvector_rotate('thc_fgs_gse_mva_mat','thc_fgs_gse',newname='mva_data_day')
        self.assertTrue((data_exists('mva_data_day')))
        self.assertTrue(data_exists('mva_lambdas'))
        self.assertTrue(data_exists('mva_min'))
        self.assertTrue(data_exists('mva_min'))
        self.assertTrue(data_exists('mva_max'))

    def test_minvar_matrix_make_hour(self):
        trange=['2007-07-10', '2007-07-11']
        pyspedas.projects.themis.fgm(probe='c', trange=trange, level='l2', coord='gse')
        minvar_matrix_make('thc_fgs_gse', twindow=3600, tslide=300)
        self.assertTrue(data_exists('thc_fgs_gse_mva_mat'))
        pyspedas.tvector_rotate('thc_fgs_gse_mva_mat','thc_fgs_gse',newname='mva_data_hour')
        self.assertTrue((data_exists('mva_data_hour')))

    def test_minvar_matrix_make_hour_interp(self):
        trange=['2007-07-10', '2007-07-11']
        pyspedas.projects.themis.fgm(probe='c', trange=trange, level='l2', coord='gse')
        minvar_matrix_make('thc_fgs_gse', twindow=3600, tslide=300)
        self.assertTrue(data_exists('thc_fgs_gse_mva_mat'))
        # Check rotation matrix input and output coords
        in_coord, out_coord = pyspedas.rotmat_get_coords("thc_fgs_gse_mva_mat")
        self.assertEqual(in_coord.upper(), "GSE")
        self.assertEqual(out_coord.upper(), "MVA")
        pyspedas.tvector_rotate('thc_fgs_gse_mva_mat','thc_fgl_gse',newname='mva_data_hour_interp')
        self.assertTrue((data_exists('mva_data_hour_interp')))
        # Check output vector coordinates
        vec_out_coord = pyspedas.get_coords('mva_data_hour_interp')
        self.assertEqual(vec_out_coord.upper(), "MVA")


    def test_minvar_matrix_make_idl_single(self):
        """ Test a minvar_matrix_make call with only a single window """
        minvar_matrix_make('idl_thb_fgs_gsm_mvaclipped1',newname='thb_fgs_gsm_mvaclipped1_mva_mat',
                           evname='mva_vals', tminname='mva_min', tmidname='mva_int', tmaxname='mva_max')

        tvector_rotate('thb_fgs_gsm_mvaclipped1_mva_mat', 'idl_thb_fgs_gsm_mvaclipped1',newname='thb_fgs_gsm_mvaclipped1_rot')
        vals = get_data('mva_vals')
        assert_allclose(vals.y, self.vals.y, rtol=1e-05, atol=1e-03)
        assert_allclose(vals.times, self.vals.times,  atol=1e-06)
        mva_mat = get_data('thb_fgs_gsm_mvaclipped1_mva_mat')
        assert_allclose(mva_mat.y, self.mat.y, rtol=1e-05, atol=1e-03)
        mva_min = get_data('mva_min')
        assert_allclose(mva_min.y, self.min.y, rtol=1e-05, atol=1e-03)
        mva_int = get_data('mva_int')
        assert_allclose(mva_int.y, self.int.y, rtol=1e-05, atol=1e-03)
        mva_max = get_data('mva_max')
        assert_allclose(mva_max.y, self.max.y, rtol=1e-05, atol=1e-03)
        mva_rot = get_data('thb_fgs_gsm_mvaclipped1_rot')
        assert_allclose(mva_rot.y, self.rot.y, rtol=1e-05, atol=1e-03)

    def test_minvar_matrix_make_idl_multi(self):
        """ Test a minvar_matrix_make call that returns multiple matrices """
        minvar_matrix_make('idl_mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar',newname='mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat',
                           twindow=16.0, tslide=4.0, evname='mva_vals2', tminname='mva_min2', tmidname='mva_int2', tmaxname='mva_max2')
        tvector_rotate('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat', 'idl_mms1_fgm_b_gsm_srvy_l2_bvec_bp',newname='mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot')
        vals = get_data('mva_vals2')
        assert_allclose(vals.times, self.vals2.times, atol=1e-06)
        assert_allclose(vals.y, self.vals2.y, rtol=1e-05, atol=1e-06)
        mva_mat = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_hp4minvar_mva_mat')
        assert_allclose(mva_mat.y, self.mat2.y, rtol=1e-03, atol=1e-05)
        mva_min = get_data('mva_min2')
        assert_allclose(mva_min.y, self.min2.y, rtol=1e-03, atol=1e-06)
        mva_int = get_data('mva_int2')
        assert_allclose(mva_int.y, self.int2.y, rtol=1e-03, atol=1e-06)
        mva_max = get_data('mva_max2')
        assert_allclose(mva_max.y, self.max2.y, rtol=1e-03, atol=1e-06)
        mva_rot = get_data('mms1_fgm_b_gsm_srvy_l2_bvec_bp_rot')
        assert_allclose(mva_rot.y, self.rot2.y, rtol=1e-05, atol=1e-06)


if __name__ == '__main__':
    unittest.main()

